package jp.topse.kaizu.sequence2csp.action;

import java.io.BufferedReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import jp.topse.kaizu.sequence2csp.convert.MergeProcessor;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.ui.IObjectActionDelegate;
import org.eclipse.ui.IWorkbenchPart;

public class AddComponentMergeAction implements IObjectActionDelegate {
	private List<IFile> files = null;

	public void setActivePart(IAction action, IWorkbenchPart targetPart) {
	}

	private void close(BufferedReader br) {
		if (br != null) {
			try {
				br.close();
			} catch (IOException ex) {
				MessageDialog.openError(null, "CSP Merge Processor", ex
						.getLocalizedMessage());
			}
		}
	}

	private void close(FileWriter fw) {
		if (fw != null) {
			try {
				fw.close();
			} catch (IOException ex) {
				MessageDialog.openError(null, "CSP Merge Processor", ex
						.getLocalizedMessage());
			}
		}
	}

	public void run(IAction action) {
		if (files != null && files.size() == 2) {
			IFile file1 = files.get(0);
			IFile file2 = files.get(1);
			BufferedReader br1 = null;
			BufferedReader br2 = null;
			FileWriter fw = null;
			try {
				br1 = new BufferedReader(new InputStreamReader(file1
						.getContents()));
				br2 = new BufferedReader(new InputStreamReader(file2
						.getContents()));

				MergeProcessor processor = new MergeProcessor("/template_merge_addcmp.txt");
				String csp = processor.merge(br1, br2);

				String filename1 = file1.getName().replaceAll("\\..*$", "");
				String filename2 = file2.getName().replaceAll("\\..*$", "");
				IFile out = file1.getParent().getFile(
						new Path(filename1 + "_" + filename2 + ".fdr2"));
				String osstring = out.getRawLocation().toOSString();
				fw = new FileWriter(osstring);
				fw.write(csp);
				fw.close();
				out.refreshLocal(IFile.DEPTH_INFINITE, null);
			} catch (IOException ex) {
				MessageDialog.openError(null, "CSP Converter", ex
						.getLocalizedMessage());
			} catch (CoreException ex) {
				MessageDialog.openError(null, "CSP Merge Processor", ex
						.getLocalizedMessage());
			} finally {
				close(br1);
				close(br2);
				close(fw);
			}
		}
	}

	public void selectionChanged(IAction action, ISelection selection) {
		files = new ArrayList<IFile>();
		if (selection instanceof IStructuredSelection) {
			IStructuredSelection structured = (IStructuredSelection) selection;
			Iterator<?> it = structured.iterator();
			while (it.hasNext()) {
				Object obj = it.next();
				if (obj instanceof IFile) {
					files.add((IFile) obj);
				}
			}
		}
	}

}
