package jp.topse.kaizu.sequence2csp.convert;

import java.io.BufferedReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import jp.topse.kaizu.sequence2csp.papyrus.Notation;

import org.eclipse.uml2.uml.BehaviorExecutionSpecification;
import org.eclipse.uml2.uml.Interaction;
import org.eclipse.uml2.uml.InteractionFragment;
import org.eclipse.uml2.uml.Lifeline;
import org.eclipse.uml2.uml.Message;
import org.eclipse.uml2.uml.MessageEnd;
import org.eclipse.uml2.uml.MessageOccurrenceSpecification;
import org.eclipse.uml2.uml.OccurrenceSpecification;
import org.eclipse.uml2.uml.StateInvariant;

public class Converter {
	public static final String CALL_KEYWORD = "call_";

	public static final String COMPO_KEYWORD = "COMPO_";

	public static final String DEFAULT_STATE_KEYWORD = "default_";

	private List<Notation> notations;

	private boolean isDebug = false;

	public Converter(List<Notation> notations) {
		this.notations = notations;
	}

	/**
	 * ̃V[PX}1FDRLqɕϊ
	 * 
	 * @param interactions
	 * @param notation
	 * @return
	 * @throws IOException
	 */
	public String convert(List<Interaction> interactions) throws IOException {
		BufferedReader br = new BufferedReader(new InputStreamReader(
				Converter.class.getResourceAsStream("/template.txt")));
		try {
			Set<Lifeline> lifelines = new HashSet<Lifeline>();
			for (Interaction interaction : interactions) {
				for (Object lifeline : interaction.getLifelines()) {
					lifelines.add((Lifeline) lifeline);
				}
			}

			Set<String> lifelineNames = new HashSet<String>();
			for (Lifeline lifeline : lifelines) {
				String name = CSPNames.get(lifeline);
				lifelineNames.add(name);
			}

			Set<String> messageNames = new HashSet<String>();
			for (Interaction interaction : interactions) {
				for (Object message : interaction.getMessages()) {
					messageNames.add(CSPNames.get((Message) message));
				}
			}

			List<InteractionFragment> fragments = new ArrayList<InteractionFragment>();
			for (Interaction interaction : interactions) {
				fragments.addAll(sortFragments(interaction.getFragments()));
			}

			String compoNames = Util.join(lifelineNames, " | ");
			String stateNames = Util.join(DEFAULT_STATE_KEYWORD,
					getLifelineStates(interactions), " | ");
			String callNames = Util.join(messageNames, " | ");
			String compoDef = convert(lifelines, fragments);

			StringBuilder sb = new StringBuilder();
			String line;
			while ((line = br.readLine()) != null) {
				line = line.replace("${CompoNames}", compoNames);
				line = line.replace("${StateNames}", stateNames);
				line = line.replace("${CallNames}", callNames);
				line = line.replace("${CompoDef}", compoDef);
				line = line.replace("${Spec}", "");
				sb.append(line);
				sb.append(Util.LINE_SEPARATOR);
			}

			return sb.toString();
		} finally {
			br.close();
		}
	}

	/**
	 * CtCɑ݂Ԗ̈ꗗԂ
	 * 
	 * @param interactions
	 * @return
	 */
	private Set<String> getLifelineStates(List<Interaction> interactions) {
		Set<String> states = new HashSet<String>();
		for (Interaction interaction : interactions) {
			for (Object fragment : interaction.getFragments()) {
				if (fragment instanceof StateInvariant) {
					StateInvariant si = (StateInvariant) fragment;
					String name = CSPNames.get(si);
					if (!states.contains(name)) {
						states.add(name);
					}
				}
			}
		}
		return states;
	}

	/**
	 * ̃V[PX}1CSPɕϊiCOMPO_̂݁j
	 * 
	 * @param lifelines
	 * @param fragments
	 * @return
	 */
	private String convert(Set<Lifeline> lifelines,
			List<InteractionFragment> fragments) {
		Map<String, SequenceTree> forest = new HashMap<String, SequenceTree>();
		for (Lifeline lifeline : lifelines) {
			convert(forest, lifeline, fragments);
		}
		StringBuilder sb = new StringBuilder();
		for (Entry<String, SequenceTree> entry : forest.entrySet()) {
			sb.append(COMPO_KEYWORD);
			sb.append("(");
			sb.append(entry.getKey());
			sb.append(") =");
			sb.append(Util.LINE_SEPARATOR);
			sb.append("\t");
			sb.append(entry.getValue().toString());
			sb.append(Util.LINE_SEPARATOR);
			sb.append(Util.LINE_SEPARATOR);
		}
		return sb.toString();
	}

	/**
	 * w肵R|[lgEԂn܂c[Ԃ
	 * 
	 * @param forest
	 * @param compoName
	 * @param startState
	 * @return
	 */
	private SequenceTree getTree(Map<String, SequenceTree> forest,
			String compoName, String startState) {
		String key = compoName + "." + startState;
		if (!forest.containsKey(key)) {
			forest.put(key, SequenceTree.create());
		}
		return forest.get(key);
	}

	/**
	 * 1{̃CtCɒڂăc[ɕϊ
	 * 
	 * @param forest
	 * @param lifeline
	 */
	private void convert(Map<String, SequenceTree> forest, Lifeline lifeline,
			List<InteractionFragment> fragments) {
		String compoName = CSPNames.get(lifeline);
		SequenceTree tree = getTree(forest, compoName, DEFAULT_STATE_KEYWORD);
		Map<OccurrenceSpecification, BehaviorExecutionSpecification> besMap = getBesMap(lifeline
				.getInteraction());

		for (int i = 0; i < fragments.size(); i++) {
			InteractionFragment fragment = (InteractionFragment) fragments
					.get(i);
			if (fragment.getCovered(lifeline.getName()) != lifeline) {
				continue;
			}

			if (fragment instanceof OccurrenceSpecification) {
				OccurrenceSpecification os = (OccurrenceSpecification) fragment;
				// bZ[WM
				if (os instanceof MessageOccurrenceSpecification) {
					tree = convert(tree, (MessageOccurrenceSpecification) os,
							lifeline);
				}
				if (besMap.containsKey(os)) {
					// ԏIc[؂
					InteractionFragment next = getNextFragment(fragments, i,
							lifeline);
					if (next instanceof StateInvariant) {
						// ̃XebvŒʏStateInvariantƂď
					} else {
						tree.addState(compoName, DEFAULT_STATE_KEYWORD);
						tree = getTree(forest, compoName, DEFAULT_STATE_KEYWORD);
					}
				}

			} else if (fragment instanceof StateInvariant) {
				StateInvariant si = (StateInvariant) fragment;
				// ԕsώ
				String stateName = CSPNames.get(si);
				tree.addState(compoName, stateName);
				tree = getTree(forest, compoName, stateName);
			}
		}
		tree.addState(compoName, DEFAULT_STATE_KEYWORD);
	}

	private List<InteractionFragment> sortFragments(
			List<InteractionFragment> list) {
		if (notations == null) {
			return list;
		}
		List<InteractionFragment> result = new ArrayList<InteractionFragment>();
		final Map<InteractionFragment, Double> map = new HashMap<InteractionFragment, Double>();
		for (InteractionFragment fragment : list) {
			map.put(fragment, 0.0);
			result.add(fragment);
			for (Notation notation : notations) {
				double y = notation.getY(fragment);
				if (y >= 0) {
					map.put(fragment, y);
					break;
				}
			}
		}
		Collections.sort(result, new Comparator<InteractionFragment>() {
			@Override
			public int compare(InteractionFragment f1, InteractionFragment f2) {
				double d = map.get(f1) - map.get(f2);
				if (d < 0) {
					return -1;
				} else if (d > 0) {
					return 1;
				} else {
					return 0;
				}
			}
		});

		if (isDebug) {
			try {
				FileWriter fw = null;
				try {
					fw = new FileWriter("sd2csp.log", true);
					PrintWriter pw = new PrintWriter(fw);
					pw.println();
					pw.println(new Date().toString());
					for (InteractionFragment f : result) {
						int y = (int) (10 * map.get(f));
						String name = f.getName();
						if (f instanceof MessageOccurrenceSpecification) {
							Message m = ((MessageOccurrenceSpecification) f)
									.getMessage();
							if (m != null) {
								name = m.getName() + "(" + f.getName() + ")";
							}
						}
						Lifeline l = f.getCovereds().get(0);
						pw.println(y + ": " + l.getName() + "(" + l.hashCode()
								+ "): " + name + ": " + f.getClass());
					}
				} finally {
					if (fw != null) {
						fw.close();
					}
				}
			} catch (IOException ex) {
				throw new RuntimeException(ex);
			}
		}
		return result;
	}

	/**
	 * bZ[WÃbZ[WŏI銈Ԃւ̃}bvB
	 * 
	 * @param interaction
	 * @return
	 */
	private Map<OccurrenceSpecification, BehaviorExecutionSpecification> getBesMap(
			Interaction interaction) {
		Map<OccurrenceSpecification, BehaviorExecutionSpecification> besMap;
		besMap = new HashMap<OccurrenceSpecification, BehaviorExecutionSpecification>();

		for (Object fragment : interaction.getFragments()) {
			if (fragment instanceof BehaviorExecutionSpecification) {
				BehaviorExecutionSpecification bes = (BehaviorExecutionSpecification) fragment;
				besMap.put(bes.getFinish(), bes);
			}
		}

		return besMap;
	}

	/**
	 * ̃CtC̎̃tOg
	 * 
	 * @param list
	 * @param prev
	 * @param lifeline
	 * @return
	 */
	private InteractionFragment getNextFragment(List<InteractionFragment> list,
			int prev, Lifeline lifeline) {
		for (int i = prev + 1; i < list.size(); i++) {
			InteractionFragment fragment = (InteractionFragment) list.get(i);
			if (fragment.getCovered(lifeline.getName()) == lifeline) {
				return fragment;
			}
		}
		return null;
	}

	/**
	 * 1̃bZ[Wc[ɕϊB
	 * 
	 * @param tree
	 * @param mos
	 * @param lifeline
	 * @return
	 */
	private SequenceTree convert(SequenceTree tree,
			MessageOccurrenceSpecification mos, Lifeline lifeline) {
		Message message = mos.getMessage();
		if (message == null) {
			// Invalid UML model
			return tree;
		}

		Lifeline sLifeline = null;
		MessageEnd sendEvent = message.getSendEvent();
		if (sendEvent instanceof InteractionFragment) {
			InteractionFragment sFragment = (InteractionFragment) sendEvent;
			sLifeline = (Lifeline) (sFragment.getCovereds().get(0));
		}

		Lifeline rLifeline = null;
		MessageEnd receiveEvent = message.getReceiveEvent();
		if (receiveEvent instanceof InteractionFragment) {
			InteractionFragment rFragment = (InteractionFragment) receiveEvent;
			rLifeline = (Lifeline) (rFragment.getCovereds().get(0));
		}

		StringBuilder sb = new StringBuilder();
		sb.append(CALL_KEYWORD);
		sb.append(".");
		sb.append(CSPNames.get(message));
		sb.append(".");
		sb.append(CSPNames.get(sLifeline));
		sb.append(".");
		sb.append(CSPNames.get(rLifeline));
		String messageString = sb.toString();
		return tree.addMessage(messageString, (rLifeline == lifeline));
	}
}
