package jp.topse.kaizu.sequence2csp.convert;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class MergeProcessor {
	private static Pattern CALL_PATTERN = Pattern
			.compile("call_\\.\\w+\\.(\\w+)\\.\\w+");

	private static String CALL_KEY = "call_";

	private String templateName;

	public MergeProcessor(String templateName) {
		this.templateName = templateName;
	}

	private <K> String getString(Map<K, String> map, K key) {
		String str = map.get(key);
		if (str != null) {
			return str;
		} else {
			return "";
		}
	}

	private Set<String> getMergedSet(String data1String, String data2String) {
		Set<String> mergedSet = new HashSet<String>();
		for (String str : data1String.split("\\|")) {
			mergedSet.add(str.trim());
		}
		for (String str : data2String.split("\\|")) {
			mergedSet.add(str.trim());
		}
		return mergedSet;
	}

	public String merge(BufferedReader br1, BufferedReader br2)
			throws IOException {
		Map<String, String> data1 = parse(br1);
		Map<String, String> data2 = parse(br2);

		BufferedReader br = null;
		try {
			br = new BufferedReader(new InputStreamReader(
					Converter.class.getResourceAsStream(templateName)));

			// CompoName_
			String compoNames1 = getString(data1, "datatype CompoName_");
			String compoNames2 = getString(data2, "datatype CompoName_");
			Set<String> compoNamesSet = new HashSet<String>();
			Set<String> compoNamesSet1 = new HashSet<String>();
			Set<String> compoNamesSet2 = new HashSet<String>();
			Set<String> sysCompoNamesSet1 = new HashSet<String>();
			Set<String> sysCompoNamesSet2 = new HashSet<String>();
			for (String str : compoNames1.split("\\|")) {
				String compo = str.trim();
				compoNamesSet.add(compo);
				compoNamesSet1.add(compo);
				if (!compo.endsWith("User")) {
					sysCompoNamesSet1.add(compo);
				}
			}
			for (String str : compoNames2.split("\\|")) {
				String compo = str.trim();
				compoNamesSet.add(compo);
				compoNamesSet2.add(compo);
				if (!compo.endsWith("User")) {
					sysCompoNamesSet2.add(compo);
				}
			}

			// StateName_
			String stateNames1 = getString(data1, "datatype StateName_");
			String stateNames2 = getString(data2, "datatype StateName_");
			Set<String> stateNamesSet = getMergedSet(stateNames1, stateNames2);

			// CallName_
			String callNames1 = getString(data1, "datatype CallName_");
			String callNames2 = getString(data2, "datatype CallName_");
			Set<String> callNamesSet = getMergedSet(callNames1, callNames2);

			// ActorName_
			String actorNames1 = getString(data1, "subtype ActorName_");
			String actorNames2 = getString(data2, "subtype ActorName_");
			Set<String> actorNamesSet = getMergedSet(actorNames1, actorNames2);

			// Component definition
			String compo1Def = getString(data1, "-- Component definition")
					.replace("COMPO_", "COMPO1_");
			String compo2Def = getString(data2, "-- Component definition")
					.replace("COMPO_", "COMPO2_");
			String compoDef = compo1Def + Util.LINE_SEPARATOR + compo2Def;

			// Internal Events
			Set<String> intenalComposSet = new HashSet<String>();
			intenalComposSet.addAll(Util.removeAll(compoNamesSet,
					compoNamesSet1));
			intenalComposSet.addAll(Util.removeAll(compoNamesSet,
					compoNamesSet2));
			String internalInterface = "interface_("
					+ Util.join(intenalComposSet, "), interface_(") + ")";
			String internalEvents;
			if (intenalComposSet.size() == 0) {
				internalEvents = "{}";
			} else if (intenalComposSet.size() == 1) {
				internalEvents = internalInterface;
			} else {
				internalEvents = "Union ({" + internalInterface + "})";
			}

			// Internal Events (abstract)
			Set<String> eventsSet = new HashSet<String>();
			Set<String> eventsSet1 = new HashSet<String>();
			Set<String> eventsSet2 = new HashSet<String>();
			for (String str : getString(data1, CALL_KEY).split("\\,")) {
				String evt = str.trim();
				eventsSet.add(evt);
				eventsSet1.add(evt);
			}
			for (String str : getString(data2, CALL_KEY).split("\\,")) {
				String evt = str.trim();
				eventsSet.add(evt);
				eventsSet2.add(evt);
			}
			Set<String> intenalEventsSet = new HashSet<String>();
			intenalEventsSet.addAll(Util.removeAll(eventsSet, eventsSet1));
			intenalEventsSet.addAll(Util.removeAll(eventsSet, eventsSet2));

			Set<String> internalEventsSetL = new HashSet<String>();
			for (String evt : intenalEventsSet) {
				Matcher m = CALL_PATTERN.matcher(evt);
				if (m.matches()) {
					if (actorNamesSet.contains(m.group(1))) {
						internalEventsSetL.add(evt);
					}
				}
			}
			Set<String> internalEventsSetE = Util.removeAll(intenalEventsSet,
					internalEventsSetL);

			String internalEventsL = "{"
					+ Util.LINE_SEPARATOR
					+ "  "
					+ Util.join(internalEventsSetL, "," + Util.LINE_SEPARATOR
							+ "  ") + Util.LINE_SEPARATOR + "}";
			String internalEventsE = "{"
					+ Util.LINE_SEPARATOR
					+ "  "
					+ Util.join(internalEventsSetE, "," + Util.LINE_SEPARATOR
							+ "  ") + Util.LINE_SEPARATOR + "}";

			// Output with Template
			StringBuilder sb = new StringBuilder();
			String line;
			while ((line = br.readLine()) != null) {
				line = line.replace("${CompoNames}",
						Util.join(compoNamesSet, " | "));
				line = line.replace("${CompoNames1}", compoNames1);
				line = line.replace("${CompoNames2}", compoNames2);
				line = line.replace("${SysCompoNames1}",
						Util.join(sysCompoNamesSet1, " | "));
				line = line.replace("${SysCompoNames2}",
						Util.join(sysCompoNamesSet2, " | "));
				line = line.replace("${StateNames}",
						Util.join(stateNamesSet, " | "));
				line = line.replace("${StateNames1}", stateNames1);
				line = line.replace("${StateNames2}", stateNames2);
				line = line.replace("${CallNames}",
						Util.join(callNamesSet, " | "));
				line = line.replace("${CallNames1}", callNames1);
				line = line.replace("${CallNames2}", callNames2);
				line = line.replace("${CompoDef}", compoDef);
				line = line.replace("${Spec}", "");
				line = line.replace("${InternalEvents}", internalEvents);
				line = line.replace("${InternalEventsE}", internalEventsE);
				line = line.replace("${InternalEventsL}", internalEventsL);
				sb.append(line);
				sb.append(Util.LINE_SEPARATOR);
			}
			return sb.toString();
		} finally {
			if (br != null) {
				br.close();
			}
		}
	}

	private Map<String, String> parse(BufferedReader br) throws IOException {
		Map<String, String> map = new HashMap<String, String>();
		Set<String> call = new HashSet<String>();
		String line;
		String key = null;
		StringBuilder value = new StringBuilder();
		while ((line = br.readLine()) != null) {
			if (line.startsWith("--")) {
				if (key != null && value.length() > 0) {
					map.put(key, value.toString());
				}
				key = line;
				value = new StringBuilder();
			} else {
				value.append(line);
				value.append(Util.LINE_SEPARATOR);

				String[] split = line.split(" = ");
				if (split.length == 2) {
					map.put(split[0], split[1]);
				}

				Matcher m = CALL_PATTERN.matcher(line);
				while (m.find()) {
					call.add(m.group());
				}
			}
		}
		map.put(CALL_KEY, Util.join(call, ","));
		return map;
	}
}
