package jp.topse.kaizu.sequence2csp.papyrus;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.eclipse.uml2.uml.ExecutionOccurrenceSpecification;
import org.eclipse.uml2.uml.ExecutionSpecification;
import org.eclipse.uml2.uml.InteractionFragment;
import org.eclipse.uml2.uml.Message;
import org.eclipse.uml2.uml.MessageOccurrenceSpecification;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public class Notation {
	private Document document;
	private Map<String, Node> elementNodeMap;
	private XPath xpath;

	public Notation(File file) throws IOException {
		try {
			DocumentBuilderFactory factory = DocumentBuilderFactory
					.newInstance();
			DocumentBuilder builder = factory.newDocumentBuilder();
			document = builder.parse(file);

			elementNodeMap = new HashMap<String, Node>();
			NodeList elements = document.getElementsByTagName("element");
			for (int i = 0; i < elements.getLength(); i++) {
				Node element = elements.item(i);
				String href = element.getAttributes().getNamedItem("href")
						.getNodeValue();
				int sharp = href.lastIndexOf('#');
				if (sharp >= 0) {
					href = href.substring(sharp + 1);
				}
				elementNodeMap.put(href, element);
			}

			xpath = XPathFactory.newInstance().newXPath();
		} catch (ParserConfigurationException ex) {
			throw new IOException(ex);
		} catch (SAXException ex) {
			throw new IOException(ex);
		}
	}

	public double getY(InteractionFragment fragment) {
		try {
			if (fragment instanceof MessageOccurrenceSpecification) {
				MessageOccurrenceSpecification mos = (MessageOccurrenceSpecification) fragment;
				Message message = mos.getMessage();
				String messageId = message.eResource().getURIFragment(message);
				if (mos == message.getSendEvent()) {
					// ԏĨbZ[WM͊ԓƂ݂Ȃ
					return getMessageY(messageId, "source") - 0.5;
				} else if (mos == message.getReceiveEvent()) {
					return getMessageY(messageId, "target");
				} else {
					return -1;
				}
			} else if (fragment instanceof ExecutionOccurrenceSpecification) {
				ExecutionOccurrenceSpecification eos = (ExecutionOccurrenceSpecification) fragment;
				ExecutionSpecification es = eos.getExecution();
				String id = es.eResource().getURIFragment(es);
				double anchor = ((eos == es.getFinish()) ? 1.0 : 0.0);
				return getY(id, anchor);
			} else {
				String id = fragment.eResource().getURIFragment(fragment);
				return getY(id, 0.0);
			}
		} catch (Exception ex) {
			return -1;
		}
	}

	public double getMessageY(String messageId, String anchorType) {
		try {
			Node elementNode = elementNodeMap.get(messageId);
			Node edgesNode = elementNode.getParentNode();
			String aid = xpath.evaluate(anchorType + "Anchor/@id", edgesNode);
			int p1 = aid.indexOf(',');
			int p2 = aid.indexOf(')', p1);
			double anchor = Double.parseDouble(aid.substring(p1 + 1, p2));

			String parentId = xpath.evaluate("@" + anchorType, edgesNode);
			String parentType = xpath.evaluate("//*[@id=\"" + parentId
					+ "\"]/element/@type", document);

			if (parentType.equals("uml:Lifeline")) {
				// Lifeline default height == 250;
				double height = getXPathDoubleValue("//*[@id=\"" + parentId
						+ "\"]/layoutConstraint/@height", document, 250);
				return height * anchor;
			} else {
				double y = getXPathDoubleValue("//*[@id=\"" + parentId
						+ "\"]/layoutConstraint/@y", document);
				double height = getXPathDoubleValue("//*[@id=\"" + parentId
						+ "\"]/layoutConstraint/@height", document);
				return y + (height * anchor);
			}
		} catch (Exception ex) {
			return -1;
		}
	}

	public double getY(String id, double anchor) {
		try {
			Node elementNode = elementNodeMap.get(id);
			Node shapeNode = elementNode.getParentNode();
			double y = getXPathDoubleValue("layoutConstraint/@y", shapeNode);
			double height = getXPathDoubleValue("layoutConstraint/@height",
					shapeNode, 0);
			return y + (height * anchor);
		} catch (Exception ex) {
			return -1;
		}
	}

	double getXPathDoubleValue(String expression, Object item)
			throws NumberFormatException, XPathExpressionException {
		return Double.parseDouble(xpath.evaluate(expression, item));
	}

	double getXPathDoubleValue(String expression, Object item, double defValue) {
		try {
			return getXPathDoubleValue(expression, item);
		} catch (Exception ex) {
			return defValue;
		}
	}
}
